// File          : ArrowLeaf3BackCap.java
// Date          : 2012-03-11
// Author        : Nicola L.C. Talbot
//                 http://theoval.cmp.uea.ac.uk/~nlct/

/*
    Copyright (C) 2006 Nicola L.C. Talbot

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

package uk.ac.uea.cmp.nlct.jdr.marker;

import java.io.*;
import java.awt.*;
import java.awt.geom.*;

import uk.ac.uea.cmp.nlct.jdr.*;

/**
 * Leaf3For cap marker. The basic marker shape looks like:
 * <img src="../images/leaf3forCapMarker.png" alt="[leaf3for shape]">
 * (for a path with pen width = 10 PostScript points).
 * This marker's shape depends on the associated path's
 * line width as well as the given marker size.
 * See {@link JDRMarker} for a description of markers.
 *
 */
public class ArrowLeaf3BackCap extends JDRMarker
{
   /**
    * Creates leaf3for marker for a path with the given pen width.
    * The marker may be repeated and/or reversed.
    * @throws InvalidRepeatValueException if repeat value is less 
    * than 1
    */
   public ArrowLeaf3BackCap(double penwidth, int repeat,
                      boolean isReversed, double arrowSize)
      throws InvalidRepeatValueException
   {
      super(penwidth, repeat, isReversed, arrowSize);

      type = ARROW_LEAF3BACK_CAP;
   }

   public String getID()
   {
      return reversed ?
           "arrow-r"+repeated+"leaf3backcap-"+size+"-"+penWidth:
           "arrow-"+repeated+"leaf3backcap-"+size+"-"+penWidth;
   }

   /**
    * Gets the path describing the basic shape of this marker.
    */
   public GeneralPath getGeneralPath()
   {
      double radius = 0.15*penWidth*size;

      double diameter = 2*radius;

      double x = radius*(1-ROOT_3);
      double y = 0.8*penWidth;

      double midX = x + radius;

      GeneralPath path = new GeneralPath();

      path.moveTo(x, 0);

      double c1 = y*ONE_OVER_ROOT_3;
      double c2 = midX-c1;

      path.curveTo(x, -c1, c2, -y, midX, -y);

      double len = 1.5*diameter;

      double endX = midX + len;

      double c3x = midX + radius*1.720465053;
      double c4x = endX - 0.6*radius;

      path.curveTo(c3x, -y,
                   c4x, 0,
                   endX, 0);

      path.curveTo(c4x, 0,
                   c3x, y,
                   midX, y);

      path.curveTo(c2, y, x, c1, x, 0);

      path.closePath();

      double shift = x;

      AffineTransform af = AffineTransform.getTranslateInstance(-shift, 0);

      Shape shape = af.createTransformedShape(path);

      Area area = new Area(shape);

      af = AffineTransform.getRotateInstance(-THREEQUARTER_PI);

      area.add(new Area(af.createTransformedShape(shape)));

      af = AffineTransform.getRotateInstance(THREEQUARTER_PI);

      area.add(new Area(af.createTransformedShape(shape)));

      af = AffineTransform.getTranslateInstance(shift, 0);

      return new GeneralPath(af.createTransformedShape(area));
   }

   public boolean isResizable()
   {
      return true;
   }

   public boolean usesLineWidth()
   {
      return true;
   }

   /**
    * Not implemented (returns empty string).
    * @deprecated
    */
   protected String pgfarrow()
   {
      return "";
   }

   public Object clone()
   {
      JDRMarker marker = null;

      try
      {
         marker = new ArrowLeaf3BackCap(penWidth, repeated,
                                         reversed, size);
         makeEqual(marker);
      }
      catch (InvalidRepeatValueException ignore)
      {
      }

      return marker;
   }

}
